/*
** File: df.h  
** Project: VxWorks Driver Framework
** Purpose: Defines API used by OS-independent driver code
**
** (C) Copyright Alpha Data 2010
*/

#if !defined(ADATA_DF_DF_H)
#define ADATA_DF_DF_H

#define DFVXW_DEBUG_USERIF   (0)
#define DFVXW_DEBUG_SPINLOCK (0)

#if !defined(DF_DBG_EXECUTION_CONTEXT)
# define DF_DBG_EXECUTION_CONTEXT (0)
#endif

#if (_WRS_VXWORKS_MAJOR <= 5)
# define DF_DECLARE_INLINE_FUNC(type, ident) static __inline__ type ident
#else
# define DF_DECLARE_INLINE_FUNC(type, ident) _WRS_INLINE type ident
#endif

#include <vxWorks.h>
#include <cacheLib.h>
#include <drv/pci/pciIntLib.h>
#include <drv/pci/pciConfigLib.h>
#include <intLib.h>
#include <iosLib.h>
#include <iv.h>
#include <limits.h>
#include <logLib.h>
#include <selectLib.h>
#include <semLib.h>
#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 6)
# define DFVXW_USE_SPINLOCKLIB (1)
# include <spinLockLib.h>
# define DFVXW_USE_VXATOMLICLIB (1)
# include <vxAtomicLib.h>
#else
# define DFVXW_USE_SPINLOCKLIB (0)
# define DFVXW_USE_VXATOMLICLIB (0)
#endif
#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
# include <stdint.h>
# define DFVXW_HAS_STDINT
#endif
#include <stdio.h>
#include <string.h>
#include <sysLib.h>
#include <taskLib.h>
#include <tickLib.h>
#include <time.h>
#include <version.h>
#include <wdLib.h>
#include <ctype.h> /* for isprint() */

#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
/* VxWorks 6.9 has some small VxWorks kernel API changes */
# define DFVXW_HAS_VX_USR_ARG_T (1)
# define DFVXW_HAS_VX_IOCTL_ARG_T (1)
# ifdef _WRS_KEEP_USING_FUNCPTR
#   define DFVXW_DRV_OPEN_RETURN_PVOID (0)
# else
#   define DFVXW_DRV_OPEN_RETURN_PVOID (1)
# endif
# define DFVXW_HAS_CACHEDMA32MALLOC (1)
#else
# define DFVXW_HAS_VX_USR_ARG_T (0)
# define DFVXW_HAS_VX_IOCTL_ARG_T (0)
# define DFVXW_DRV_OPEN_RETURN_PVOID (0)
# define DFVXW_HAS_CACHEDMA32MALLOC (0)
#endif

#if defined(ADB3_VXBUS)
# if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 9)
#   define DFVXW_VXBUS5 (1)
#   define DFVXW_VXBUS_DMA (1)
# elif (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 7)
#   define DFVXW_VXBUS4 (1)
# elif (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 6)
#   define DFVXW_VXBUS3 (1)
# else
#   error "The VxBus ADB3 driver can be built only for VxWorks 6.6 or later. For a legacy (non-VxBus driver), add TYPE=legacy to the make command line."
# endif
# include <vxBusLib.h>
# include <hwif/util/hwMemLib.h>
# include <hwif/util/vxbDmaBufLib.h>
# include <hwif/vxbus/vxBus.h>
# include <../src/hwif/h/vxbus/vxbAccess.h>
# define DFVXW_USE_ISRDEFERLIB (0)
# define DFVXW_USE_EXCLIB      (1)
# include <isrDeferLib.h>
#else
# include <drv/pci/pciIntLib.h>
# include <drv/pci/pciConfigLib.h>
# define DFVXW_USE_ISRDEFERLIB (0)
# define DFVXW_USE_EXCLIB      (1)
#endif

#define DF_FMT_I64 "%lld"
#define DF_FMT_U64 "%llu"
#define DF_FMT_H64 "%llx"
#define DF_FMT_16H64 "%016llx"

#if DFVXW_USE_VXATOMLICLIB
# define dfVxwBarrier() VX_MEM_BARRIER_RW()
#else
/* Before VxWorks 6.6, there was no SMP support, so memory barriers not needed */
# define dfVxwBarrier()
#endif

/* Atomic integer type */
#if DFVXW_USE_VXATOMLICLIB

/* Use vxAtomicLib for atomic integer type */
typedef atomicVal_t DfAtomicInt;

#define dfAtomicSet(p, value)  ((void)vxAtomicSet(p, (atomicVal_t)value))
#define dfAtomicInc(p)         ((int)vxAtomicInc(p) + 1)
#define dfAtomicDec(p)         ((int)vxAtomicDec(p) - 1)
#define dfAtomicGet(p, value)  ((int)vxAtomicGet(p))

#else

/* Fall back to using intLock() & intUnlock() */
typedef int DfAtomicInt;

DF_DECLARE_INLINE_FUNC(void, dfAtomicSet)(
  DfAtomicInt* p,
  int newValue)
{
  *p = newValue;
}

DF_DECLARE_INLINE_FUNC(int, dfAtomicGet)(
  DfAtomicInt* p)
{
  return *p;
}

DF_DECLARE_INLINE_FUNC(int, dfAtomicInc)(
  DfAtomicInt* p)
{
  int flags, newValue;

  flags = intLock();
  newValue = ++(*p);
  intUnlock(flags);

  return newValue;
}

DF_DECLARE_INLINE_FUNC(int, dfAtomicDec)(
  DfAtomicInt* p)
{
  int flags, newValue;

  flags = intLock();
  newValue = --(*p);
  intUnlock(flags);

  return newValue;
}

#endif

#define DF_DBG_BUILD 1

typedef int boolean_t;

#ifndef DFVXW_HAS_STDINT
typedef unsigned long uintptr_t;
#endif

#define DF_HAS_USER_EVENT_HANDLE (1)
typedef SEM_ID user_event_handle_t;

#include <../common/dfcommon.h>
#include <../common/dflist.h>

/* Endian conversion */

#if defined(_BYTE_ORDER)
# if (_BYTE_ORDER == _BIG_ENDIAN)

/* Big-endian byte order */

DF_DECLARE_INLINE_FUNC(uint64_t, dfCpuToLe64)(
  uint64_t x)
{
  return ((x        & (uint64_t)0x000000ff) << 56) | ((x        & (uint64_t)0x0000ff00) << 40) | ((x        & (uint64_t)0x00ff0000) << 24) | ((x        & (uint64_t)0xff000000) <<  8) |
         ((x >>  8) & (uint64_t)0xff000000)        | ((x >> 24) & (uint64_t)0x00ff0000)        | ((x >> 40) & (uint64_t)0x0000ff00)        | ((x >> 56) & (uint64_t)0x000000ff);
}

DF_DECLARE_INLINE_FUNC(uint32_t, dfCpuToLe32)(
  uint32_t x)
{
  return ((x & (uint32_t)0x00ff) << 24) | ((x & (uint32_t)0xff00) << 8) | ((x >> 8) & (uint32_t)0xff00) | ((x >> 24) & (uint32_t)0x00ff);
}

DF_DECLARE_INLINE_FUNC(uint16_t, dfCpuToLe16)(
  uint16_t x)
{
  return ((x & (uint16_t)0xff) << 8) | ((x >> 8) & (uint16_t)0xff);
}

#   define dfCpuToLe8(x)  (x)
#   define dfLe64ToCpu(x) dfCpuToLe64(x)
#   define dfLe32ToCpu(x) dfCpuToLe32(x)
#   define dfLe16ToCpu(x) dfCpuToLe16(x)
#   define dfLe8ToCpu(x)  (x)

# else /* end of (_BYTE_ORDER == _BIG_ENDIAN) */

/* Little-endian byte order */

#   define dfCpuToLe64(x) (x)
#   define dfCpuToLe32(x) (x)
#   define dfCpuToLe16(x) (x)
#   define dfCpuToLe8(x)  (x)
#   define dfLe64ToCpu(x) (x)
#   define dfLe32ToCpu(x) (x)
#   define dfLe16ToCpu(x) (x)
#   define dfLe8ToCpu(x)  (x)

# endif /* end of !(_BYTE_ORDER == _BIG_ENDIAN) */
#else

# error Macro _BYTE_ORDER is not defined. Cannot detect endianness of this architecture.

#endif

/* 64-bit division */

DF_DECLARE_INLINE_FUNC(uint64_t, dfDivide64By32)(
  uint64_t dividend,
  uint32_t divisor)
{
  return dividend / divisor;
}

DF_DECLARE_INLINE_FUNC(uint64_t, dfDivide64By32Rem)(
  uint64_t dividend,
  uint32_t divisor,
  uint32_t* pRemainder)
{
  *pRemainder = (uint32_t)(dividend % divisor);
  return dividend / divisor;
}

/* This is nasty :( but convenient */
#define dfSplitUint64(x) (unsigned long)(uint32_t)((x) >> 32U), (unsigned long)(uint32_t)(x)

DF_DECLARE_INLINE_FUNC(uint16_t, dfReadUnaligned16)(
	uint16_t* p)
{
	uint8_t* p8 = (uint8_t*)p;
	uint16_t val16;

	val16 = (uint16_t)(((uint16_t)*p8++) << 0U);
	val16 = (uint16_t)(val16 | (((uint16_t)*p8) << 8U));

	return dfLe16ToCpu(val16);
}

DF_DECLARE_INLINE_FUNC(uint32_t, dfReadUnaligned32)(
	uint32_t* p)
{
	uint8_t* p8 = (uint8_t*)p;
	uint32_t val32;

	val32  = (uint32_t)(((uint32_t)*p8++) << 0U);
	val32 |= (uint32_t)(((uint32_t)*p8++) << 8U);
	val32 |= (uint32_t)(((uint32_t)*p8++) << 16U);
	val32 |= (uint32_t)(((uint32_t)*p8) << 24U);

	return dfLe32ToCpu(val32);
}

DF_DECLARE_INLINE_FUNC(uint64_t, dfReadUnaligned64)(
	uint64_t* p)
{
	uint8_t* p8 = (uint8_t*)p;
	uint64_t val64;

	val64  = ((uint64_t)*p8++) << 0U;
	val64 |= ((uint64_t)*p8++) << 8U;
	val64 |= ((uint64_t)*p8++) << 16U;
	val64 |= ((uint64_t)*p8++) << 24U;
	val64 |= ((uint64_t)*p8++) << 32U;
	val64 |= ((uint64_t)*p8++) << 40U;
	val64 |= ((uint64_t)*p8++) << 48U;
	val64 |= ((uint64_t)*p8) << 56U;

	return dfLe64ToCpu(val64);
}

/*
** Attributes for functions. It would be nice if we had some kind of
** tool able parse the code and use these attributes to look for errors.
*/
#define DF_ATTR_PASSIVE_CONTEXT /* Must be called in a thread context */

/* Return value from DfOnIoctlMethod functions */
typedef int DfIoStatus;
#define DfIoStatusSuccess      (0)                           /* Completed successfully */
#define DfIoStatusNoMemory     (ENOMEM)                      /* Couldn't allocate memory for data structures */
#define DfIoStatusAccessDenied (EACCES)                      /* Insufficient privileges */
#define DfIoStatusInvalid      (EFAULT)                      /* pIoctl, inSize and/or outSize were invalid */
#define DfIoStatusInvalidCode  (ENOSYS)                      /* IOCTL code was invald */
#define DfIoStatusCancelled    (ECANCELED)                   /* Cancelled */
#define DfIoStatusDeferred     (EINPROGRESS)                 /* Will be completed later */
#define DfIoStatusError(x)     ((x) | 0x40000000)            /* Defines an API-specific error */
#define DfIoStatusIsError(x)   ((x) != DfIoStatusSuccess && (x) != DfIoStatusDeferred)
#define DfIoStatusIsUser(x)    (((x) & 0x40000000) != 0)

/*
** Forward declarations of structures and types exported by framework
*/

struct _DfBufferDescription;
typedef struct _DfBufferDescription DfBufferDescription;

struct _DfBusResources;
typedef struct _DfBusResources DfBusResources;

struct _DfClientObject;
typedef struct _DfClientObject DfClientObject;

struct _DfClientRequest;
typedef struct _DfClientRequest DfClientRequest;

typedef void DfCancelCallback(DfClientRequest* pRequest, DfIoStatus reason);

struct _DfDeviceObject;
typedef struct _DfDeviceObject DfDeviceObject;

struct _DfDmaBuffer;
typedef struct _DfDmaBuffer DfDmaBuffer;

struct _DfDmaMapperNode;
typedef struct _DfDmaMapperNode DfDmaMapperNode;

struct _DfDmaMapperPosition;
typedef struct _DfDmaMapperPosition DfDmaMapperPosition;

struct _DfDmaMapperSlice;
typedef struct _DfDmaMapperSlice DfDmaMapperSlice;

typedef void DfDmaMapperCallback(DfDeviceObject* pDevObj, unsigned int mapIndex, DfDmaMapperNode* pTable, unsigned int tableLength, DfDmaMapperPosition* pProgress, void* pContext);

struct _DfDpc;
typedef struct _DfDpc DfDpc;

typedef void DfDpcFunction(DfDpc* pDpc, void* pContext, void* pArg);

struct _DfDriverObject;
typedef struct _DfDriverObject DfDriverObject;

struct _DfEvent;
typedef struct _DfEvent DfEvent;

struct _DfMutex;
typedef struct _DfMutex DfMutex;

struct _DfExportedInterface;
typedef struct _DfExportedInterface DfExportedInterface;

struct _DfImportedInterface;
typedef struct _DfImportedInterface DfImportedInterface;

struct _DfInterface;
typedef struct _DfInterface DfInterface;

struct _DfInterfaceClass;
typedef struct _DfInterfaceClass DfInterfaceClass;

struct _DfInterfaceMethods;
typedef struct _DfInterfaceMethods DfInterfaceMethods;

struct _DfInterruptObject;
typedef struct _DfInterruptObject DfInterruptObject;

typedef boolean_t DfIsr(DfInterruptObject* pInterruptObj, void* pContext);

struct _DfMappedIoRegion;
typedef struct _DfMappedIoRegion DfMappedIoRegion;

typedef void* DfMemoryHandle;

/* Encapsulates the notion of an address space and a pointer in that address space. */
#if defined(ADB3_VXBUS)
/* VxBus driver */
typedef struct {
  void* h;    /* Handle used to ensure that VxBus subsystem uses correct method of access */
  uint8_t* p; /* Register address whose meaning depends on pHandle */
} DfRegPointer;
#else
/* Legacy VxWorks driver */
typedef struct {
  uint8_t* p;
} DfRegPtr;
#endif

struct _DfMmappableRegion;
typedef struct _DfMmappableRegion DfMmappableRegion;

struct _DfMmappedRegion;
typedef struct _DfMmappedRegion DfMmappedRegion;

struct _DfPool;
typedef struct _DfPool DfPool;

#if DFVXW_USE_SPINLOCKLIB

typedef struct _DfSpinLock {
  spinlockIsr_t lock;
# if DFVXW_DEBUG_SPINLOCK
  unsigned int count;
# endif
} DfSpinLock;

typedef unsigned int DfSpinLockFlags;

#else

typedef struct _DfVxwSpinLockDummy {
# if DFVXW_DEBUG_SPINLOCK
  unsigned int count;
# else
  unsigned int dummy;
# endif
} DfSpinLock;

typedef int DfSpinLockFlags;

#endif

struct _DfTimer;
typedef struct _DfTimer DfTimer;

typedef SEM_ID DfUserEvent;

#include <../common/dfentry.h>

/*
** Device structures
*/

struct _DfPool {
  struct {
    DfSpinLock lock;
    void* pBuffer;
    void** pFreeStack;
    unsigned int stackPtr;
    unsigned int numItem;
#if DF_DBG_BUILD
    size_t itemSize;
    size_t itemSizeAligned;
#endif
  } system;
};

/* Nonzero to use semaphore pools for dfEvent{Init|Uninit} and dfMutex{Init|Uninit} instead of always allocating */
#define DFVXW_EVENT_USE_SEMPOOL (1)

struct _DfEvent {
  struct {
    SEM_ID semId;
#if DFVXW_EVENT_USE_SEMPOOL
    boolean_t bFromPool;
#endif
  } system;
};

struct _DfMutex {
  struct {
    SEM_ID semId;
#if DFVXW_EVENT_USE_SEMPOOL
    boolean_t bFromPool;
#endif
  } system;
};

struct _DfClientRequest {
  struct {
    DfClientObject* pClObj;
    DfCancelCallback* pCancelCallback;
    boolean_t bCancel;
    boolean_t bCompleted;
    DfIoStatus status;
    DfEvent* pWaitEvent;
    DfListNode node;
    void* pContext; /* Client code can get/set value via dfRequest{Get|Set}Context */
  } system;
};

#define DF_PCIBARF_IO       (0x1U)
#define DF_PCIBARF_PREFETCH (0x8U)
struct _DfPciBar {
	uint64_t base;
	uint64_t size;
	uint32_t flags;
};
typedef struct _DfPciBar DfPciBar;

/* This datatype must be treated as being opaque */
struct _DfPciIrqLine {
  boolean_t bMessageSignalled;
  boolean_t bLevelSensitive;
  union {
    struct {
      uint8_t intPin;
      uint8_t intLine;
    } raw;
#if defined(ADB3_VXBUS)
    /* Nothing yet */
#else
    struct { /* This struct must be considered opaque */
      uint8_t intPin;
      uint8_t intLine;
    } translated;
#endif
  } variant;
};
typedef struct _DfPciIrqLine DfPciIrqLine;

#define DF_PCIBUSRESOURCESF_EXPROMVALID (0x2U)
struct _DfPciBusResources {
	uint32_t flags;
	unsigned int numMemBar;
	DfPciBar memBar[6];
	DfPciBar expRomBar;
	unsigned int numIoBar;
	DfPciBar ioBar[6];
	unsigned int numIrq;
	DfPciIrqLine irq[4];
};
typedef struct _DfPciBusResources DfPciBusResources;

struct _DfBusResources {
	union {
		DfPciBusResources pci;
	} variant;
};

struct _DfPciHardwareId {
	uint16_t vendor, device;
	uint16_t subVendor, subDevice;
	uint8_t revision;
};
typedef struct _DfPciHardwareId DfPciHardwareId;

#define DF_PCI_MATCH_VENDOR    (0x1U << 0)
#define DF_PCI_MATCH_DEVICE    (0x1U << 1)
#define DF_PCI_MATCH_SUBVENDOR (0x1U << 2)
#define DF_PCI_MATCH_SUBDEVICE (0x1U << 3)
#define DF_PCI_MATCH_REVISION  (0x1U << 4)

#define DF_PCI_COMPATIBLE_ID_END { { 0U, 0U, 0U, 0U, 0U }, 0xffU }

struct _DfPciCompatibleId {
  DfPciHardwareId id;
  uint16_t match;
};
typedef struct _DfPciCompatibleId DfPciCompatibleId;

struct _DfUnknownHardwareId {
    char id[48];
};
typedef struct _DfUnknownHardwareId DfUnknownHardwareId;

typedef enum _DfBusType {
  DfBusInvalid = 0,
  DfBusUnknown = 1,
  DfBusPci = 2
} DfBusType;

struct _DfHardwareId {
	DfBusType busType;
	union {
    DfUnknownHardwareId unknown;
		DfPciHardwareId pci;
	} variant;
};
typedef struct _DfHardwareId DfHardwareId;

struct _DfCompatibleId {
	DfBusType busType;
	union {
		DfPciCompatibleId pci;
	} variant;
};
typedef struct _DfCompatibleId DfCompatibleId;

struct _DfVxwClientRequestNonBlock;
typedef struct _DfVxwClientRequestNonBlock DfVxwClientRequestNonBlock;

struct _DfClientObject {
  struct { /* Client code must not touch this struct */
    DfSpinLock lock;
#if DFVXW_DEBUG_USERIF
    unsigned int entered; /* Counts number of threads currently executing an IOCTL */
#endif
    DfDeviceObject* pDevObj;
    DfInterface* pInterface;
    DfMmappedRegion* pMmappedRegions;
    struct {
      /* List of client requests that have a cancel callback */
      unsigned int count;
      DfList list;
    } requests;
    struct {
      /* Information about non-blocking client requests */
      int state;
      DfVxwClientRequestNonBlock* pHead; /* Currently, at most one allowed at any moment */
      SEL_WAKEUP_LIST selWakeupList;     /* Used to handle FIOSELECT / FIOUNSELECT */
      DfEvent closeEvent;
    } nonblock;
    void* pContext; /* Client code can get/set value via dfClient{Get|Set}Context */
  } system;
};

DF_DECLARE_INLINE_FUNC(void*, dfClientGetContext)(
 DfClientObject* pClObj)
{
  return pClObj->system.pContext;
}

DF_DECLARE_INLINE_FUNC(void, dfClientSetContext)(
 DfClientObject* pClObj,
 void* pContext)
{
  pClObj->system.pContext = pContext;
}

struct _DfVxwDmaMapperContext;
typedef struct _DfVxwDmaMapperContext DfVxwDmaMapperContext;

struct _DfDeviceObject {
  struct { /* Client code must not touch this struct */
#if defined(ADB3_VXBUS)
    /* For VxBus-compliant driver, must be first element of this struct */
    VXB_DEVICE_ID pVxbDev;
#else
    DfDeviceObject* pNext;
#endif
    struct {
      unsigned int maxMapRegister;
      size_t maxMapSize;
      unsigned int numMap;
      DfVxwDmaMapperContext* pMapArray;
    } dma;
    DfExportedInterface* pExportedInterfaces;
    DfImportedInterface* pImportedInterfaces;
    DfMappedIoRegion* pMappedIoRegions;
    DfMmappableRegion* pMmappableRegions;
    struct {
      DfInterface* pHead;
    } interfaceList;
    struct {
      DfPool pool;
      DfPool poolNonBlock;
      boolean_t bPoolValid;
      boolean_t bPoolNonBlockValid;
    } request;
    /* Client code can get/set any value here via dfDeviceGetContext / dfDeviceSetContext */
    void* pContext;
  } system;

  DfDriverObject* pDrvObj;
  DfHardwareId hardwareId;
};

DF_DECLARE_INLINE_FUNC(void, dfDeviceSetContext)(
  DfDeviceObject* pDevObj,
  void* pContext)
{
  pDevObj->system.pContext = pContext;
}

DF_DECLARE_INLINE_FUNC(void*, dfDeviceGetContext)(
  DfDeviceObject* pDevObj)
{
  return pDevObj->system.pContext;
}

/*
** Driver structures
*/

typedef boolean_t
DfOnAddMethod(
	DfDeviceObject* pDevObj);

typedef void
DfOnRemoveMethod(
	DfDeviceObject* pDevObj);

typedef boolean_t
DfOnStartMethod(
	DfDeviceObject* pDevObj,
	DfBusResources* pRawResources,
	DfBusResources* pTranslatedResources);

typedef void
DfOnStopMethod(
	DfDeviceObject* pDevObj,
  boolean_t bCanTouchHardware);

typedef boolean_t /* bVeto */
DfOnPowerQueryMethod(
	DfDeviceObject* pDevObj,
  unsigned int currentState,
  unsigned int requestedState,
  boolean_t bIsShutdown);

typedef void
DfOnPowerSetMethod(
	DfDeviceObject* pDevObj,
  unsigned int currentState,
  unsigned int newState,
  boolean_t bIsShutdown);

typedef DfIoStatus
DfOnOpenMethod(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj,
  DfClientRequest* pReq,
  DfInterface* pInterface,
  boolean_t bPrivileged);

typedef void
DfOnCleanupMethod(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj);

typedef void
DfOnCloseMethod(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj);

typedef DfIoStatus
DfOnIoctlMethod(
	DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfClientRequest* pReq,
  unsigned int ioctlCode,
  void* pIoctl,
  unsigned int inSize,
  unsigned int outSize);

typedef enum _DfQueryMapStatus {
  DfQueryMapSuccess = 0,
  DfQueryMapInvalidIndex,
  DfQueryMapInvalidRegion,
  DfQueryMapGeneralFailure 
} DfQueryMapStatus;

struct _DfDriverObject {
  unsigned int deviceContextSize;

  struct {
    DfOnAddMethod* pOnAdd;
    DfOnRemoveMethod* pOnRemove;
    DfOnStartMethod* pOnStart;
    DfOnStopMethod* pOnStop;
    DfOnPowerQueryMethod* pOnPowerQuery;
    DfOnPowerSetMethod* pOnPowerSet;
  } methods;

#if !defined(ADB3_VXBUS)
  /* Private */
  struct {
    struct {
      DfDeviceObject* pHead;
      DfDeviceObject* pTail;
    } devices;
  } system;
#endif
};

/*
** Debug messages
*/

#if DF_DBG_BUILD
extern int g_dfDebugLevel;

extern void
_dfDebugPrint(
	const char* fmt, ...);

# define dfDebugLevel(LEVEL) ((LEVEL) <= g_dfDebugLevel)

# define dfDebugPrint(LEVEL, ARGS) \
do { if ((LEVEL) <= g_dfDebugLevel) _dfDebugPrint ARGS; } while (0)

# define dfDebugTrace(LEVEL) \
do { if (dfDebugLevel(LEVEL)) _dfDebugPrint("reached %s:%u\n", __FILE__, __LINE__); } while (0)

extern void
dfDebugDumpMemory(
  int level,
	unsigned int indent,
	unsigned int nColumn,
	unsigned int radix,
	boolean_t bAlign,
	const void* pBuffer,
	size_t n);

# define dfAssert(expr) do { if (!(expr)) _dfDebugPrint("*** %s:%u assertion failed: %s\n", __FILE__, __LINE__, #expr); } while (0)

# define dfBugCheck(ARGS) _dfDebugPrint ARGS;

#else

# define dfAssert(expr)
# define dfBugCheck(ARGS)
# define dfDebugDumpMemory(LEVEL, INDENT, NCOLUMN, RADIX, BALIGN, PBUFFER, N)
# define dfDebugPrint(LEVEL, ARGS)
# define dfDebugLevel(LEVEL) FALSE

#endif

#if DF_DBG_BUILD && DF_DBG_EXECUTION_CONTEXT

# define dfAssertPassiveContext()     /* TO DO */
# define dfAssertNotHoldingSpinLock() /* TO DO */

#else

# define dfAssertPassiveContext()
# define dfAssertNotHoldingSpinLock()

#endif

/*
** Import and export of direct-call interfaces
*/

#define DF_MAX_INTERFACE_SIZE (1024)
extern boolean_t
dfImportInterface(
	DfDeviceObject* pDevObj,
	const char* pName,
	unsigned int size,
	void* pInterfaceStruct,
  void** ppInterfaceContext);

extern boolean_t
dfExportInterface(
	DfDeviceObject* pDevObj,
	const char* pName,
	unsigned int size,
	void* pInterfaceStruct,
  void* pInterfaceContext);

extern boolean_t
dfReleaseInterface(
	DfDeviceObject* pDevObj,
	const char* pName);

extern boolean_t
dfUnexportInterface(
	DfDeviceObject* pDevObj,
	const char* pName);

/*
** Mapping and unmapping I/O space for access by driver (NOT user space)
*/

extern void*
dfIoSpaceMap(
  DfDeviceObject* pDevObj,
  DfBusResources* pResources,
  unsigned int barIndex,
  uint64_t offset,
  uint64_t length,
  DfMemoryHandle* pHandle);

extern boolean_t
dfIoSpaceUnmap(
	DfDeviceObject* pDevObj,
	void* pKernelAddress);

/*
** Memory space register access
*/

/*
** PCI Memory register access functions; the user can assume that they perform
** conversion to little-endian endian format if necessary.
*/

#if defined(ADB3_VXBUS)

DF_DECLARE_INLINE_FUNC(void, dfPciMemWrite32)(
  DfMemoryHandle handle,
  uint32_t* p,
  uint32_t d)
{
  vxbWrite32(handle, (UINT32*)p, d);
}

DF_DECLARE_INLINE_FUNC(uint32_t, dfPciMemRead32)(
  DfMemoryHandle handle,
  uint32_t* p)
{
  return vxbRead32(handle, (UINT32*)p);
}

DF_DECLARE_INLINE_FUNC(void, dfPciMemWrite16)(
  DfMemoryHandle handle,
  uint16_t* p,
  uint16_t d)
{
  vxbWrite16(handle, p, d);
}

DF_DECLARE_INLINE_FUNC(uint16_t, dfPciMemRead16)(
  DfMemoryHandle handle,
  uint16_t* p)
{
  return vxbRead16(handle, p);
}

DF_DECLARE_INLINE_FUNC(void, dfPciMemWrite8)(
  DfMemoryHandle handle,
  uint8_t* p,
  uint8_t d)
{
  vxbWrite8(handle, p, d);
}

DF_DECLARE_INLINE_FUNC(uint8_t, dfPciMemRead8)(
  DfMemoryHandle handle,
  uint8_t* p)
{
  return vxbRead8(handle, p);
}

#else

DF_DECLARE_INLINE_FUNC(void, dfPciMemWrite32)(
  DfMemoryHandle handle, /* ignored */
  uint32_t* p,
  uint32_t d)
{
  volatile uint32_t* vp = (volatile uint32_t*)p;

  *vp = dfCpuToLe32(d);
  dfVxwBarrier();
}

DF_DECLARE_INLINE_FUNC(uint32_t, dfPciMemRead32)(
  DfMemoryHandle handle, /* ignored */
  uint32_t* p)
{
  uint32_t val32 = *p;

  return dfLe32ToCpu(val32);
}

DF_DECLARE_INLINE_FUNC(void, dfPciMemWrite16)(
  DfMemoryHandle handle, /* ignored */
  uint16_t* p,
  uint16_t d)
{
  volatile uint16_t* vp = (volatile uint16_t*)p;

  *vp = dfCpuToLe16(d);
  dfVxwBarrier();
}

DF_DECLARE_INLINE_FUNC(uint16_t, dfPciMemRead16)(
  DfMemoryHandle handle, /* ignored */
  uint16_t* p)
{
  uint16_t val16 = *p;

  return dfLe16ToCpu(val16);
}

DF_DECLARE_INLINE_FUNC(void, dfPciMemWrite8)(
  DfMemoryHandle handle, /* ignored */
  uint8_t* p,
  uint8_t d)
{
  volatile uint8_t* vp = (volatile uint8_t*)p;

  *vp = dfCpuToLe8(d);
  dfVxwBarrier();
}

DF_DECLARE_INLINE_FUNC(uint8_t, dfPciMemRead8)(
  DfMemoryHandle handle, /* ignored */
  uint8_t* p)
{
  uint8_t val8 = *p;

  return dfLe8ToCpu(val8);
}

#endif

DF_DECLARE_INLINE_FUNC(void, dfMemWrite32)(
  DfMemoryHandle handle, /* ignored */
  uint32_t* p,
  uint32_t d)
{
  *(volatile uint32_t*)p = d;
  dfVxwBarrier();
}

DF_DECLARE_INLINE_FUNC(uint32_t, dfMemRead32)(
  DfMemoryHandle handle, /* ignored */
  uint32_t* p)
{
  return *p;
}

DF_DECLARE_INLINE_FUNC(void, dfMemWrite16)(
  DfMemoryHandle handle, /* ignored */
  uint16_t* p,
  uint16_t d)
{
  *(volatile uint16_t*)p = d;
  dfVxwBarrier();
}

DF_DECLARE_INLINE_FUNC(uint16_t, dfMemRead16)(
  DfMemoryHandle handle, /* ignored */
  uint16_t* p)
{
  return *p;
}

DF_DECLARE_INLINE_FUNC(void, dfMemWrite8)(
  DfMemoryHandle handle, /* ignored */
  uint8_t* p,
  uint8_t d)
{
  *(volatile uint8_t*)p = dfCpuToLe8(d);
  dfVxwBarrier();
}

DF_DECLARE_INLINE_FUNC(uint8_t, dfMemRead8)(
  DfMemoryHandle handle, /* ignored */
  uint8_t* p)
{
  return *p;
}

/*
** Critical sections and mutual exclusion
**
** Uses spinLockLib if VxWorks 6.6 or later, otherwise intLock() / intUnlock().
*/

DF_DECLARE_INLINE_FUNC(void, dfSpinLockInit)(
  DfSpinLock* pLock)
{
#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrInit(&pLock->lock, 0);
#endif
#if DFVXW_DEBUG_SPINLOCK
  pLock->count = 0;
#endif
}

#if DFVXW_USE_SPINLOCKLIB

DF_DECLARE_INLINE_FUNC(DfSpinLockFlags, dfSpinLockGet)(
  DfSpinLock* pLock)
{
# if defined(SPIN_LOCK_TRACE)
  spinLockIsrTake(&pLock->lock, __FILE__, __LINE__);
# else
  spinLockIsrTake(&pLock->lock);
# endif
# if DFVXW_DEBUG_SPINLOCK
  dfAssert(pLock->count == 0);
  pLock->count++;
# endif
  return 0;
}

#else

DF_DECLARE_INLINE_FUNC(DfSpinLockFlags, dfSpinLockGet)(
  DfSpinLock* pLock)
{
  DfSpinLockFlags flags = intLock();
# if DFVXW_DEBUG_SPINLOCK
  dfAssert(pLock->count == 0);
  pLock->count++;
# endif
  return flags;
}

#endif

DF_DECLARE_INLINE_FUNC(void, dfSpinLockPut)(
  DfSpinLock* pLock,
  DfSpinLockFlags flags)
{
# if DFVXW_DEBUG_SPINLOCK
  dfAssert(pLock->count == 1);
  pLock->count--;
#endif
#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrGive(&pLock->lock);
#else
  intUnlock(flags);
#endif
}

/*
** Delaying execution
*/

#if (_WRS_VXWORKS_MAJOR >= 6)

/* VxWorks 6.x has guaranteed 64-bit tick counter, which eliminates tick counter wrapping problems */
typedef uint64_t DfTime;

# define dfTimeGet tick64Get

# define dfTimeGreaterEqual(a, b) ((a) >= (b))
# define dfTimeGreater(a, b) ((a) > (b))

#else

/* VxWorks 5.x and earlier doesn't have guaranteed 64-bit tick counter, so fall back to using tickGet() */
typedef ULONG DfTime;
# define dfTimeGet tickGet

/* This won't work correctly for intervals greater than or equal to (ULONG_MAX+1)/2 ticks */
# define dfTimeGreaterEqual(a, b) ((long)((a) - (b)) >= 0)
# define dfTimeGreater(a, b) ((long)((a) - (b)) > 0)

#endif

# define dfTimeLesserEqual(a, b) (!dfTimeGreater(a, b))
# define dfTimeLesser(a, b) (!dfTimeGreaterEqual(a, b))

/* Delays CPU for specified number of microseconds */
/* DANGER - this may block the calling task in VxWorks, so currently CANNOT be used inside a spinlock critical section. */
DF_DECLARE_INLINE_FUNC(void, dfDelayMicroseconds)(
  unsigned int us)
{
  struct timespec ts;

  dfAssert(us < 1000000);
  ts.tv_sec = 0;
  ts.tv_nsec = us * 1000;
  nanosleep(&ts, NULL);
}

DF_DECLARE_INLINE_FUNC(DfTime, dfMicrosecondsToTime)(
  unsigned int us)
{
  const uint32_t M = 1000000U;
  uint64_t tmp;

  tmp = (uint64_t)us * (unsigned int)sysClkRateGet();
  return (DfTime)((tmp + M - 1) / M);
}

DF_DECLARE_INLINE_FUNC(boolean_t, dfDelayThreadUntil(DfTime when)) {
  DfTime now, delta;
  int n;

  now = dfTimeGet();
  while (when < now) {
    delta = when - now;
    n = (delta >= INT_MAX) ? INT_MAX : (int)delta;
    taskDelay(n);
    now = dfTimeGet();
  }
  return FALSE; /* TRUE => wait was interrupted */
}

DF_DECLARE_INLINE_FUNC(boolean_t, dfDelayThreadFor)(
  DfTime delay)
{
  return dfDelayThreadUntil(dfTimeGet() + delay); /* TRUE if wait was interrupted */
}

/*
** Kernel event waiting and signalling
*/

#if DFVXW_EVENT_USE_SEMPOOL
extern SEM_ID g_dfVxwSemBPoolArray[];
extern unsigned int g_dfVxwSemBPoolPtr;
extern unsigned int g_dfVxwSemBPoolSize;
extern DfSpinLock g_dfVxwSemBPoolLock;
#endif

DF_DECLARE_INLINE_FUNC(void, dfEventInit)(
  DfEvent* pEvent)
{
#if DFVXW_EVENT_USE_SEMPOOL
  DfSpinLockFlags f;

  dfAssertPassiveContext();

  /* Get a semaphore from the semaphore pool */
  f = dfSpinLockGet(&g_dfVxwSemBPoolLock);
  if (g_dfVxwSemBPoolPtr < g_dfVxwSemBPoolSize) {
    pEvent->system.semId = g_dfVxwSemBPoolArray[g_dfVxwSemBPoolPtr];
    g_dfVxwSemBPoolPtr++;
    dfSpinLockPut(&g_dfVxwSemBPoolLock, f);
    pEvent->system.bFromPool = TRUE;
  } else {
    dfSpinLockPut(&g_dfVxwSemBPoolLock, f);
    /* Fall back to allocating dynamically */
    pEvent->system.semId = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
    pEvent->system.bFromPool = FALSE;
  }
#else
  dfAssertPassiveContext();

  pEvent->system.semId = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfEventUninit)(
  DfEvent* pEvent)
{
#if DFVXW_EVENT_USE_SEMPOOL
  DfSpinLockFlags f;

  dfAssertPassiveContext();

  /* Return the semaphore to the semaphore pool */
  if (pEvent->system.bFromPool) {
    /* Ensure that the semaphore is empty */
    semTake(pEvent->system.semId, 0);
    f = dfSpinLockGet(&g_dfVxwSemBPoolLock);
    g_dfVxwSemBPoolPtr--;
    g_dfVxwSemBPoolArray[g_dfVxwSemBPoolPtr] = pEvent->system.semId;
    dfSpinLockPut(&g_dfVxwSemBPoolLock, f);
  } else {
    semDelete(pEvent->system.semId);
  }
#else
  dfAssertPassiveContext();

  semDelete(pEvent->system.semId);
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfEventWait)(
  DfEvent* pEvent)
{
  dfAssertPassiveContext();

  semTake(pEvent->system.semId, WAIT_FOREVER);
}

DF_DECLARE_INLINE_FUNC(boolean_t, dfEventWaitInterruptible)(
  DfEvent* pEvent)
{
  dfAssertPassiveContext();

  dfEventWait(pEvent);
  return FALSE;
}

DF_DECLARE_INLINE_FUNC(void, dfEventSignal)(
  DfEvent* pEvent)
{
  dfAssertNotHoldingSpinLock();

#if (_WRS_VXWORKS_MAJOR > 6) || (_WRS_VXWORKS_MAJOR == 6 && _WRS_VXWORKS_MINOR >= 7)
  semBGiveScalable(pEvent->system.semId, SEM_NO_ID_VALIDATE);
#else
  semGive(pEvent->system.semId);
#endif
}

/*
** Kernel mutex
*/

#if DFVXW_EVENT_USE_SEMPOOL
extern SEM_ID g_dfVxwSemMPoolArray[];
extern unsigned int g_dfVxwSemMPoolPtr;
extern unsigned int g_dfVxwSemMPoolSize;
extern DfSpinLock g_dfVxwSemMPoolLock;
#endif

DF_DECLARE_INLINE_FUNC(void, dfMutexInit)(
  DfMutex* pMutex)
{
#if DFVXW_EVENT_USE_SEMPOOL
  DfSpinLockFlags f;

  /* Get a semaphore from the semaphore pool */
  f = dfSpinLockGet(&g_dfVxwSemMPoolLock);
  if (g_dfVxwSemMPoolPtr < g_dfVxwSemMPoolSize) {
    pMutex->system.semId = g_dfVxwSemMPoolArray[g_dfVxwSemMPoolPtr];
    g_dfVxwSemMPoolPtr++;
    dfSpinLockPut(&g_dfVxwSemMPoolLock, f);
    pMutex->system.bFromPool = TRUE;
  } else {
    dfSpinLockPut(&g_dfVxwSemMPoolLock, f);
    /* Fall back to allocating dynamically */
    pMutex->system.semId = semMCreate(SEM_Q_FIFO);
    pMutex->system.bFromPool = FALSE;
  }
#else
  pMutex->system.semId = semMCreate(SEM_Q_FIFO);
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfMutexUninit)(
  DfMutex* pMutex)
{
#if DFVXW_EVENT_USE_SEMPOOL
  DfSpinLockFlags f;

  /* Return the semaphore to the semaphore pool */
  if (pMutex->system.bFromPool) {
    /* Ensure that the semaphore is empty */
    semTake(pMutex->system.semId, 0);
    f = dfSpinLockGet(&g_dfVxwSemMPoolLock);
    g_dfVxwSemMPoolPtr--;
    g_dfVxwSemMPoolArray[g_dfVxwSemMPoolPtr] = pMutex->system.semId;
    dfSpinLockPut(&g_dfVxwSemMPoolLock, f);
  } else {
    semDelete(pMutex->system.semId);
  }
#else
  semDelete(pMutex->system.semId);
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfMutexAcquire)(
  DfMutex* pMutex)
{
  semTake(pMutex->system.semId, WAIT_FOREVER);
}

DF_DECLARE_INLINE_FUNC(boolean_t, dfMutexAcquireInterruptible)(
  DfMutex* pMutex)
{
  return (semTake(pMutex->system.semId, WAIT_FOREVER) != OK) ? TRUE : FALSE;
}

DF_DECLARE_INLINE_FUNC(void, dfMutexRelease)(
  DfMutex* pMutex)
{
  semGive(pMutex->system.semId);
}

/*
** Timers
*/

typedef void DfTimerCallback(DfTimer* pTimer, void* pTimerContext, void* pCallbackContext);

void
_dfVxwTimerStub(
#if DFVXW_HAS_VX_USR_ARG_T
  _Vx_usr_arg_t arg);
#else
  int arg);
#endif

struct _DfTimer {
  struct {
    DfTime expires;
    DfTimerCallback* pCallback;
    void* pTimerContext;
    void* pCallbackContext;
    boolean_t bFinalDelay;
    WDOG_ID wdogId;
  } system;
};

extern void
dfTimerInit(
  DfTimer* pDfTimer,
  void* pTimerContext);

extern void
dfTimerUninit(
  DfTimer* pDfTimer);

DF_DECLARE_INLINE_FUNC(void, dfTimerSchedule)(
  DfTimer* pDfTimer,
  DfTime expires,
  DfTimerCallback* pCallback,
  void* pCallbackContext)
{
  STATUS result;
  DfTime now, delayRemaining;
  int delay;

  pDfTimer->system.pCallback = pCallback;
  pDfTimer->system.pCallbackContext = pCallbackContext;
  pDfTimer->system.expires = expires;
  now = dfTimeGet();
  if (expires < now) {
    pDfTimer->system.bFinalDelay = TRUE;
    delay = 0;
  } else {
    delayRemaining = expires - now;
    if (delayRemaining > INT_MAX) {
      delay = INT_MAX;
      pDfTimer->system.bFinalDelay = FALSE;
    } else {
      delay = (int)delayRemaining;
      pDfTimer->system.bFinalDelay = TRUE;
    }
  }
#if DFVXW_HAS_VX_USR_ARG_T
  result = wdStart(pDfTimer->system.wdogId, delay, (FUNCPTR)_dfVxwTimerStub, (_Vx_usr_arg_t)pDfTimer);
#else
  result = wdStart(pDfTimer->system.wdogId, delay, (FUNCPTR)_dfVxwTimerStub, (int)pDfTimer);
#endif
  dfAssert(OK == result);
}

DF_DECLARE_INLINE_FUNC(boolean_t, dfTimerCancel)(
  DfTimer* pDfTimer)
{
  return OK == wdCancel(pDfTimer->system.wdogId) ? TRUE : FALSE;
}

/*
** Registering interfaces to user-mode applications
*/

struct _DfInterfaceClass {
  struct {
    char name[32];
    int drvNum;
    unsigned int numInterface;
  } system;
};

struct _DfInterfaceMethods {
  DfOnCleanupMethod* pOnCleanup;
  DfOnCloseMethod* pOnClose;
  DfOnIoctlMethod* pOnIoctl;
#if DF_NEED_THUNK
  DfOnIoctlMethod* pOnIoctlThunk;
#endif
  DfOnOpenMethod* pOnOpen;
};

struct _DfInterface {
  struct {
    /* This member MUST be the first member of this struct! */
    DEV_HDR iosDev;
    char name[32];
    DfInterface* pNext;
    DfDeviceObject* pDevObj;
    DfInterfaceClass* pClass;
    boolean_t bEnabled;
    DfInterfaceMethods methods;
    unsigned int index;
    /* Client code can get/set this value via DfInterface{Get|Set}Context */
    void* pContext;
  } system;
};

extern boolean_t
dfInterfaceClassRegister(
  DfDriverObject* pDrvObj,
  const char* pClassName,
  DfInterfaceClass** ppClass);

extern void
dfInterfaceClassUnregister(
  DfInterfaceClass* pClass);

extern boolean_t
dfInterfaceRegister(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface,
  DfInterfaceClass* pClass,
  const DfInterfaceMethods* pMethods);

extern boolean_t
dfInterfaceUnregister(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface);

extern boolean_t
dfInterfaceEnable(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface,
  boolean_t bEnable);

DF_DECLARE_INLINE_FUNC(void*, dfInterfaceGetContext)(
  DfInterface* pInterface)
{
  return pInterface->system.pContext;
}

DF_DECLARE_INLINE_FUNC(void, dfInterfaceSetContext)(
  DfInterface* pInterface,
  void* pContext)
{
  pInterface->system.pContext = pContext;
}

/*
** Handling of I/O requests
*/

/* Returns TRUE if cancel routine was cleared OK, otherwise FALSE */
extern boolean_t
dfRequestClearCallback(
  DfClientRequest* pClientRequest);

extern void
dfRequestComplete(
  DfClientRequest* pClientRequest,
  DfIoStatus status);

/* Returns TRUE if cancel routine was set OK, otherwise FALSE */
extern boolean_t
dfRequestSetCallback(
  DfClientRequest* pClientRequest,
  DfCancelCallback* pCallback);

DF_DECLARE_INLINE_FUNC(void*, dfRequestGetContext)(
  DfClientRequest* pClientRequest)
{
  return pClientRequest->system.pContext;
}

DF_DECLARE_INLINE_FUNC(void, dfRequestSetContext)(
  DfClientRequest* pClientRequest,
  void* pContext)
{
  pClientRequest->system.pContext = pContext;
}

#define dfRequestPending(pClientRequest) /* Do nothing */

/*
** Accessing driver parameters
*/

typedef enum _DfParameterStatus {
    DfParameterSuccess = 0,     /* Success */
    DfParameterNotFound,        /* A parameter named 'pKeyName' was not found */
    DfParameterNoMemory,        /* Couldn't allocate memory complete operation */
    DfParameterInvalid,         /* Arguments were invalid */
    DfParameterGeneralFailure   /* All other (unexpected) failures */
} DfParameterStatus;

extern DfParameterStatus
dfParameterGetUint32(
    DfDriverObject* pDrvObj,
    const char* pKeyName,
    uint32_t* pValue);

extern DfParameterStatus
dfParameterGetBoolean(
    DfDriverObject* pDrvObj,
    const char* pKeyName,
    boolean_t* pValue);

/*
** pBufferNeeded   bufferSize    pBufferNeeded          Purpose
** -------------------------------------------------------------------------------------------------------------------------------
** NULL            0             NULL                   tests if parameter exists
** NULL            0             non-NULL               tests if parameter exists & returns buffer size required
** NULL            >0            (any)                  (illegal, status always DfParameterInvalid)
** non-NULL        0             (any)                  (illegal, status always DfParameterInvalid)
** non-NULL        >0            NULL                   tests if parameter exists & returns parameter value
** non-NULL        >0            non-NULL               tests if parameter exists & returns parameter value + buffer size required
*/
extern DfParameterStatus
dfParameterGetString(
    DfDriverObject* pDrvObj,
    const char* pKeyName,
    size_t bufferSize,
    size_t* pBufferNeeded,
    char* pBuffer);

/* Allocates buffer to hold value; caller's responsibility to free using dfFree */
extern DfParameterStatus
dfParameterGetStringAlloc(
    DfDriverObject* pDrvObj,
    const char* pKeyName,
    char** ppBuffer);

/*
** Allocating and freeing memory
*/

#if defined(DF_DBG_BUILD) && defined(DF_TRACK_MALLOC)

#define dfMalloc(n) _dfVxwDbgMalloc(n, __FILE__, __LINE__)
#define dfFree(p) _dfVxwDbgFree(p, __FILE__, __LINE__)

extern void*
_dfVxwDbgMalloc(
  size_t n,
  const char* pFilename,
  int lineNum);

extern void
_dfVxwDbgFree(
  void* p,
  const char* pFilename,
  int lineNum);

#else

DF_DECLARE_INLINE_FUNC(void*, dfMalloc)(
  size_t n)
{
  /* TO DO - this needs modifications to use boot mem pool for VxBus? */
  void* p;

  p = malloc(n);
  if (NULL == p) {
    return NULL;
  }
  bzero(p, n);
  return p;
}

# define dfFree(p) free(p)

#endif

#define dfMallocThread(n) dfMalloc(n)
#define dfFreeThread(p)   dfFree(p)

/*
** String handling
*/

DF_DECLARE_INLINE_FUNC(void, dfStrCpy)(
  char* pDst,
  size_t charSize,
  const char* pSrc)
{
  char c;

  if (0 == charSize) {
    return;
  }
  charSize--;
  while (charSize) {
    c = *pSrc++;
    if (c == '\0') {
      break;
    }
    *pDst++ = c;
    charSize--;
  }
  *pDst = '\0';
}

DF_DECLARE_INLINE_FUNC(void, dfStrCat)(
  char* pDst,
  size_t charSize,
  const char* pSrc)
{
  char c;

  while (charSize) {
    c = *pDst;
    if (c == '\0') {
      break;
    }
    charSize--;
    pDst++;
  }
  dfStrCpy(pDst, charSize, pSrc);
}

#define dfSPrintF snprintf

/* 
** Data copying
*/

#define dfZeroMemory(pDst, nBytes) bzero((char*)(pDst), nBytes);

#define dfSetMemory8(pDst, nBytes, val8) memset(pDst, (int)val8, nBytes);

#define DF_ZERO_OBJECT(pDst) do { \
  dfAssert(NULL != (pDst)); \
  bzero((char*)(pDst), sizeof(*(pDst))); \
} while (0)

#define DF_SET_OBJECT8(pDst, val8) do { \
  dfAssert(NULL != (pDst)); \
  dfSetMemory8(pDst, sizeof(*(pDst)), val8); \
} while (0)

/* Kernel to kernel memory copy */
#define dfCopyKK(pDst, pSrc, n) bcopy((const char*)(pSrc), (char*)(pDst), n)

#define DF_USER_SPACE_BEGIN(unique) /* do nothing */
#define DF_USER_SPACE_EXCEPT(unique) goto _dfBufferAccessSkip##unique;
#define DF_USER_SPACE_END(unique) _dfBufferAccessSkip##unique:

/* Must be enclosed by DF_USER_SPACE_BEGIN ... DF_USER_SPACE_EXCEPT */
/* User-space to I/O memory copy */
#define dfCopyUI(pDst, pSrc, n, unique) bcopy((const char*)(pSrc), (char*)(pDst), n)

/* Must be enclosed by DF_USER_SPACE_BEGIN ... DF_USER_SPACE_EXCEPT */
/* I/O memory to user-space copy */
#define dfCopyIU(pDst, pSrc, n, unique) bcopy((const char*)(pSrc), (char*)(pDst), n)

/* Must be enclosed by DF_USER_SPACE_BEGIN ... DF_USER_SPACE_EXCEPT */
/* User-space to kernel memory copy */
#define dfCopyUK(pDst, pSrc, n, unique) bcopy((const char*)(pSrc), (char*)(pDst), n)

/* Must be enclosed by DF_USER_SPACE_BEGIN ... DF_USER_SPACE_EXCEPT */
/* Kernel to user-space memory copy */
#define dfCopyKU(pDst, pSrc, n, unique) bcopy((const char*)(pSrc), (char*)(pDst), n)

/* Must be enclosed by DF_USER_SPACE_BEGIN ... DF_USER_SPACE_EXCEPT */
#define dfUserBufferValidate(pBuffer, length, bWrite, unique) /* do nothing */

/*
** Buffer descriptions
*/

typedef enum _DfDescriptionResult {
  DfDescriptionSuccess = 0,
  DfDescriptionUnexpectedError = 1,
  DfDescriptionNullPointer = 2,
  DfDescriptionInvalidBuffer = 3,
  DfDescriptionNoMemory = 4
} DfDescriptionResult;

struct _DfBufferDescription {
  const void* pAddress;
  size_t length;

  /* The members of the following array are private */
  struct {
    boolean_t bUserMode;
    boolean_t bLocked;
  } system;
};

extern DfDescriptionResult
dfBufferDescriptionCreateUser(
  const void* pUserAddress,
  size_t length,
  DfBufferDescription** ppDescription);

extern void
dfBufferDescriptionDelete(
  DfBufferDescription* pDescription);

extern boolean_t
dfBufferDescriptionLock(
  DfBufferDescription* pDescription);

extern void
dfBufferDescriptionUnlock(
  DfBufferDescription* pDescription);

/*
** Registering and unregistering user-space mappable regions
*/

typedef enum _DfMmappableStatus {
  DfMmappableSuccess = 0,     /* Success */
  DfMmappableNoMemory,        /* Couldn't allocate memory complete operation */
  DfMmappableInvalidRegion,   /* Arguments were invalid */
  DfMmappableRegionTooLarge,  /* Region is too large */
  DfMmappableGeneralFailure   /* All other (unexpected) failures */
} DfMmappableStatus;

extern DfMmappableStatus
dfMmappableRegionRegister(
	DfDeviceObject* pDevObj,
  uint32_t regionTag,
  boolean_t bIsMemory,
  void* pKernelBase,
  size_t kernelSize,
  uint64_t translatedBase,  /* ignored if bIsMemory is TRUE */
  uint64_t translatedSize); /* ignored if bIsMemory is TRUE */

extern DfMmappableStatus
dfMmappableRegionUnregister(
	DfDeviceObject* pDevObj,
  uint32_t regionTag);

extern DfMmappableRegion*
dfMmappableRegionFind(
	DfDeviceObject* pDevObj,
  unsigned int regionTag);

/*
** Pools
*/

#define dfPoolInit(pPool, pDevObj, name, type) _dfPoolInit(pPool, pDevObj, name, sizeof(type))

#define dfPoolAlloc(pPool, type) ((type*)_dfPoolAlloc(pPool))

extern boolean_t
_dfPoolInit(
  DfPool* pPool,
  DfDeviceObject* pDevObj, /* ignored */
  const char* pName, /* ignored */
  size_t objectSize);

extern void
dfPoolUninit(
  DfPool* pPool);

DF_DECLARE_INLINE_FUNC(void*, _dfPoolAlloc)(
  DfPool* pPool)
{
  DfSpinLockFlags f;
  unsigned int numItem, stackPtr;
  void* pItem;

  numItem = pPool->system.numItem;
  f = dfSpinLockGet(&pPool->system.lock);
  stackPtr = pPool->system.stackPtr;
  if (stackPtr >= numItem) {
    dfSpinLockPut(&pPool->system.lock, f);
    return NULL;
  }
  pItem = pPool->system.pFreeStack[stackPtr];
  pPool->system.stackPtr = stackPtr + 1;
  dfSpinLockPut(&pPool->system.lock, f);

  return pItem;
}

DF_DECLARE_INLINE_FUNC(void, dfPoolFree)(
  DfPool* pPool,
  void* pItem)
{
  DfSpinLockFlags f;
  unsigned int stackPtr;

  f = dfSpinLockGet(&pPool->system.lock);
  stackPtr = pPool->system.stackPtr;
  dfAssert(stackPtr != 0);
  stackPtr--;
  pPool->system.pFreeStack[stackPtr] = pItem;
  pPool->system.stackPtr = stackPtr;
  dfSpinLockPut(&pPool->system.lock, f);
}

/*
** Interrupt handling
*/

struct _DfInterruptObject {
  struct {
#if DFVXW_USE_SPINLOCKLIB
    spinlockIsr_t lock;
#endif
    boolean_t bConnected;
    DfIsr* pIsr;
    void* pContext;
#if defined(ADB3_VXBUS)
    struct {
      unsigned int index;
    } vxbus;
#else
    struct {
      VOIDFUNCPTR* irqVector;
      int irqNumber;
    } legacy;
#endif
#if DFVXW_USE_ISRDEFERLIB
    ISR_DEFER_QUEUE_ID isrDeferQueueId;
#endif
  } system;
};

extern boolean_t
dfInterruptConnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject,
  DfBusResources* pBusResources,
  DfBusType busType,
  unsigned int interruptIndex,
  DfIsr* pIsr,
  void* pContext);

extern void
dfInterruptDisconnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject);

extern boolean_t
dfInterruptUsingMsi(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject);

DF_DECLARE_INLINE_FUNC(DfSpinLockFlags, dfInterruptSpinLockGet)(
  DfInterruptObject* pInterruptObject)
{
#if DFVXW_USE_SPINLOCKLIB
# if defined(SPIN_LOCK_TRACE)
  spinLockIsrTake(&pInterruptObject->system.lock, __FILE__, __LINE__);
# else
  spinLockIsrTake(&pInterruptObject->system.lock);
# endif
  return 0;
#else
  return (DfSpinLockFlags)intLock();
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfInterruptSpinLockPut)(
  DfInterruptObject* pInterruptObject,
  DfSpinLockFlags flags)
{
#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrGive(&pInterruptObject->system.lock);
#else
  intUnlock((int)flags);
#endif
}

DF_DECLARE_INLINE_FUNC(DfSpinLockFlags, dfIsrSpinLockGet)(
  DfInterruptObject* pInterruptObject)
{
#if DFVXW_USE_SPINLOCKLIB
# if defined(SPIN_LOCK_TRACE)
  spinLockIsrTake(&pInterruptObject->system.lock, __FILE__, __LINE__);
# else
  spinLockIsrTake(&pInterruptObject->system.lock);
# endif
  return 0;
#else
  /* Must not intLock in interrupt context */
  return 0;
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfIsrSpinLockPut)(
  DfInterruptObject* pInterruptObject,
  DfSpinLockFlags flags)
{
#if DFVXW_USE_SPINLOCKLIB
  spinLockIsrGive(&pInterruptObject->system.lock);
#else
  /* Must not intUnlock in interrupt context */
#endif
}

/*
** User-mode event handling
*/

DF_DECLARE_INLINE_FUNC(boolean_t, dfUserEventGet)(
  user_event_handle_t hUserEvent,
  DfUserEvent* pUserEvent)
{
  if (NULL == (DfUserEvent*)hUserEvent) {
    return FALSE;
  }
  *pUserEvent = hUserEvent;
  return TRUE;
}

#define dfUserEventPut(userEvent) /* Nothing to do */

DF_DECLARE_INLINE_FUNC(void, dfUserEventSignal)(
  DfUserEvent userEvent)
{
  STATUS status;

  status = semGive(userEvent);
  dfAssert(OK == status);
}

/*
** Deferred procedure calls
*/

extern void _dfVxwIsrDeferStub(void* pContext);
#if DFVXW_HAS_VX_USR_ARG_T
extern void _dfVxwExcJobStub(_Vx_usr_arg_t arg1, _Vx_usr_arg_t arg2, _Vx_usr_arg_t arg3, _Vx_usr_arg_t arg4, _Vx_usr_arg_t arg5, _Vx_usr_arg_t arg6);
#else
extern void _dfVxwExcJobStub(int arg1, int arg2, int arg3, int arg4, int arg5, int arg6);
#endif

struct _DfDpc {
  struct {
#if DFVXW_USE_ISRDEFERLIB
    ISR_DEFER_JOB isrDeferJob;
    void* pArg;
    DfInterruptObject* pInterruptObject;
    DfAtomicInt refCount;
    DfEvent* pEvent;
#elif DFVXW_USE_EXCLIB
    void* pArg;
    DfAtomicInt refCount;
    DfEvent* pEvent;
#endif
    DfDpcFunction* pFunction;
    void* pContext;
  } system;
};

DF_DECLARE_INLINE_FUNC(void, dfDpcInit)(
  DfDpc* pDpc,
  DfInterruptObject* pInterruptObject,
  DfDpcFunction* pFunction,
  void* pContext)
{
#if DFVXW_USE_ISRDEFERLIB
  pDpc->system.pInterruptObject `= pInterruptObject;
#endif
  pDpc->system.pFunction = pFunction;
  pDpc->system.pContext = pContext;
  dfAtomicSet(&pDpc->system.refCount, 1);
}

DF_DECLARE_INLINE_FUNC(void, dfDpcSchedule)(
  DfDpc* pDpc,
  void* pArg)
{
#if DFVXW_USE_ISRDEFERLIB
  /* Use isrDeferLib (VxBus drivers only) */
  DfInterruptObject* pInterruptObject = pDpc->system.pInterruptObject;

  dfAssert(intContext());
  dfAssert(NULL != pInterruptObject);
  DfAtomicInc(&pDpc->system.refCount);
  pDpc->system.isrDeferJob.func = _dfVxwIsrDeferStub;
  pDpc->system.isrDeferJob.pData = (void*)(uintptr_t)pDpc;
  pDpc->system.pArg = pArg;
  isrDeferJobAdd(pInterruptObject->system.isrDeferQueueId, &pDpc->system.isrDeferJob);
#elif DFVXW_USE_EXCLIB
  /* Use Exception Task */
  int result;

  (void)dfAtomicInc(&pDpc->system.refCount);
#if DFVXW_HAS_VX_USR_ARG_T
  result = excJobAdd(_dfVxwExcJobStub, (_Vx_usr_arg_t)pDpc, (_Vx_usr_arg_t)pArg, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0, (_Vx_usr_arg_t)0);
#else
  result = excJobAdd(_dfVxwExcJobStub, (int)pDpc, (int)pArg, 0, 0, 0, 0);
#endif
  dfAssert(OK == result);
#else
  /* Fall back to calling function directly */
  pDpc->system.pFunction(pDpc, pDpc->system.pContext, pArg);
#endif
}

DF_DECLARE_INLINE_FUNC(void, dfDpcUninit)(
  DfDpc* pDpc)
{
#if DFVXW_USE_ISRDEFERLIB || DFVXW_USE_EXCLIB
  DfEvent wait;
  int refCount;

  dfAssertPassiveContext();
  
  dfEventInit(&wait);
  pDpc->system.pEvent = &wait;
  refCount = dfAtomicDec(&pDpc->system.refCount);
  if (refCount > 0) {
    dfEventWait(&wait);
  }
  dfEventUninit(&wait);
#else
  /* Nothing to do, as dfDpcSchedule invokes the callback directly */
  dfAssertPassiveContext();
#endif
}

/*
** DMA support
**
** dfDmaMapperAllocate
**   - Declares to the system that this driver requires a DMA mapper.
**   - Must not be called twice for the same device object without an intervening call to
**     dfDmaMapperFree.
**
** dfDmaMapperFree
**   - Inverse of dfDmaMapperAllocate; undoes a call to dfDmaMapperAllocate.
**   - Must be called after dfDmaMapperAllocate.
**
** dfDmaBufferAllocate
**   - Allocates a small persistent buffer for holding data structures such as a DMA engine linked list.
**   - Must be called after dfDmaMapperAllocate & before dfDmaMapperFree.
**   - Requested block size must be <= g_dfDmaBufferMaxLength.
**
** dfDmaBufferFree
**   - Inverse of dfDmaBufferAllocate; undoes a call to dfDmaBufferAllocate.
**   - Must be called after dfDmaBufferAllocate & before dfDmaMapperFree.
**
** dfDmaMapperSetup
** dfDmaMapperSetupThread
**   - Gets a scatter-gather table for a specified buffer description.
**   - Must be called after dfDmaMapperAllocate & before dfDmaMapperFree.
**   - Calls a client-supplied callback function when the scatter-gather tables is ready; may cause
**     resources such as bounce-buffers or hardware scatter-gather table entries to be allocated.
**   - dfDmaMapperSetupThread is the version that must be called when executing in the context of a
**     thread, as opposed to DPC context.
**
** dfDmaMapperTeardown
**   - Inverse of dfDmaMapperSetup; undoes a call to dfDmaMapperSetup.
**   - Must be called after dfDmaMapperSetup & before dfDmaMapperFree.
**   - Frees any resources such as bounce-buffers or hardware scatter-gather table entries.
*/

#define g_dfDmaBufferMaxLength (0x10000U)

struct _DfDmaMapperPosition {
  size_t sliceIndex;
  size_t slicePosition;
};

struct _DfDmaBuffer {
  uint64_t busAddress;
  void* pKernelAddress;
  size_t size;

  struct { /* Private */
    DfDeviceObject* pDevObj;
#if defined(ADB3_VXBUS)
    struct {
      VXB_DMA_TAG_ID tagId;
      VXB_DMA_MAP_ID mapId;
    } vxbus;
#endif
    struct {
      struct {
        void* pKernelAddress;
        size_t size;
      } unaligned;
    } legacy;
  } system;
};

struct _DfDmaMapperSlice {
  size_t offset;
  size_t length;
  uint64_t localAddress;
};

struct _DfDmaMapperNode {
  uint64_t busAddress;
  uint64_t localAddress;
  size_t length;
  struct {
    void* pKernelAddress;
  } system;
};

/* Must be called after dfDmaMapperAllocate */
extern DfDmaBuffer*
dfDmaBufferAllocate(
  DfDeviceObject* pDevObj,
  size_t size,
  size_t alignment);

/* Must be called after dfDmaBufferAllocate */
extern void
dfDmaBufferFree(
  DfDmaBuffer* pDmaBuffer);

extern boolean_t
dfDmaMapperAllocate(
  DfDeviceObject* pDevObj,
  unsigned int numberOfMaps,
  unsigned int mapLength,          /* In blocks */
  boolean_t b64BitAddressing,
  size_t maxMapSize,
  unsigned int* pActualMapLength); /* In blocks */

/* Must be called after dfDmaMapperAllocate */
extern void
dfDmaMapperFree(
  DfDeviceObject* pDevObj);

/* Must be called after dfDmaMapperAllocate */
extern boolean_t
dfDmaMapperSetup(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex,
  DfBufferDescription* pBufferDescription,
  DfDmaMapperCallback* pCallback,
  void* pCallbackContext,
  boolean_t bWriteToDevice,
  boolean_t bFixedLocalAddress,
  DfDmaMapperSlice* pSlices,
  size_t numSlice,
  DfDmaMapperPosition* pStartPosition);

/* Must be called after dfDmaMapperAllocate */
#define dfDmaMapperSetupThread dfDmaMapperSetup

/* Must be called after dfDmaMapperSetup */
extern void
dfDmaMapperTeardown(
  DfDeviceObject* pDevObj,
  unsigned int mapIndex);

/*
** System sleeping disable / enable
**
** dfSystemSleepEnable
**   - Hints to the system that we don't want the system to sleep when idle.
**
** dfSystemSleepReenable - allows system to sleep again
**   - Hints to the system that it's OK again for system to sleep when idle.
*/

DF_DECLARE_INLINE_FUNC(void*, dfSystemSleepDisable)(
  DfDeviceObject* pDevObj /* ignored */)
{
  /* Currently does nothing in VxWorks */
  return (void*)(uintptr_t)1U;
}

DF_DECLARE_INLINE_FUNC(void, dfSystemSleepReenable)(
  DfDeviceObject* pDevObj, /* ignored */
  void* hDisableSleep)
{
  /* Currently does nothing in VxWorks */
  dfAssert(NULL != hDisableSleep);
}

#endif
