/*
** File: dfvxworks.h  
** Project: VxWorks Driver Framework
** Purpose: Defines API used by VxWorks-dependent driver code
**
** (C) Copyright Alpha Data 2010
*/

#if !defined(ADATA_DF_DFVXWORKS_H)
#define ADATA_DF_DFVXWORKS_H

typedef struct _DfVxwDriverParameterUint32 {
  const char* pName;
  unsigned int* pValue;
} DfVxwDriverParameterUint32;

typedef struct _DfVxwDriverParameterBoolean {
  const char* pName;
  unsigned int* pValue;
} DfVxwDriverParameterBoolean;

typedef struct _DfVxwDriverParameterString {
  const char* pName;
  char* pValue;
  size_t length;
} DfVxwDriverParameterString;

#define DFVXW_MAX_IOCTL_SIZE (64)

struct _DfVxwClientRequestNonBlock {
  struct {
    uint8_t data[DFVXW_MAX_IOCTL_SIZE];
  } ioctlStruct;
  DfClientRequest clientRequest;
};

struct _DfMmappableRegion {
  struct {
    DfMmappableRegion* pNext;
    uint32_t regionTag;
    void* pKernelBase;
    size_t kernelSize;
  } system;
};

struct _DfMmappedRegion {
  struct {
  	DfMmappedRegion* pNext;
	  void* pKernelBase;
	  size_t kernelSize;
	  void* pUserVirtualBase;
  } system;
};

#if defined(ADB3_VXBUS)
/* Maximum number of bytes per VxBus DMA map */
# define DFVXW_VXBUS_DMA_MAX_SEGMENT_LENGTH (0x1000)
#endif

struct _DfVxwDmaMapperContext {
  DfDeviceObject* pDeviceObject;
  unsigned int index;
  DfDmaMapperNode* pTable;
#if defined(ADB3_VXBUS)
  struct {
    VXB_DMA_TAG_ID tagId;
    VXB_DMA_MAP_ID* pMapIdTable;
  } vxbus;
#endif

  struct {
    unsigned int numEntry;
    boolean_t bWriteToDevice;
#if DF_DBG_BUILD
    boolean_t bIdle;
#endif
  } curr;
};

struct _DfExportedInterface {
  DfDeviceObject* pDevObj;
  char name[32];
  unsigned int size;
  void* pInterfaceStruct;
  void* pInterfaceContext;
  DfExportedInterface* pNext;
};

struct _DfImportedInterface {
  char name[32];
  unsigned int size;
  DfImportedInterface* pNext;
};

extern STATUS
dfVxbEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware);

extern void
dfVxbExit(
  void);

extern STATUS
dfVxwLegacyEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware);

extern void
dfVxwLegacyExit(
  void);

extern const char g_dfDriverName[];

extern DfVxwDriverParameterUint32  g_dfVxwDriverParamUint32[];
extern DfVxwDriverParameterBoolean g_dfVxwDriverParamBoolean[];
extern DfVxwDriverParameterString  g_dfVxwDriverParamString[];

/* Translate a bus address to the corresponding CPU address; TRUE => success */
extern boolean_t 
dfVxwLegacyBusToCpuAddress(
  DfDeviceObject* pDevObj,
  uint64_t busAddress,
  void** pKernelAddress);

/* Translate a CPU address to the corresponding bus address; TRUE => success */
extern boolean_t
dfVxwLegacyCpuToBusAddress(
  DfDeviceObject* pDevObj,
  void* pKernelAddress,
  uint64_t* pBusAddress);

/* Sync the data cache before or after a DMA transfer */
extern void
dfVxwLegacyCacheSync(
  void* pKernelAddress,
  boolean_t bWriteToDevice,
  size_t n);

extern boolean_t
dfVxwLegacyInterruptConnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject,
  DfBusResources* pBusResources,
  DfBusType busType,
  unsigned int interruptIndex,
  DfIsr* pIsr,
  void* pContext);

extern void
dfVxwLegacyInterruptDisconnect(
  DfDeviceObject* pDevObj,
  DfInterruptObject* pInterruptObject);

#if DF_DBG_BUILD

extern void
dfVxwLogInit(void);

extern void
dfVxwLogUninit(void);

#else

# define dfVxwLogInit()   /* Do nothing */
# define dfVxwLogUninit() /* Do nothing */

#endif

#if DFVXW_EVENT_USE_SEMPOOL
extern boolean_t
dfVxwSemPoolInit(void);

extern void
dfVxwSemPoolUninit(void);
#endif

extern DfMmappableStatus
dfVxwMmapRegion(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj,
  uint32_t regionTag,
  uint64_t offset,
  uint64_t size,
  void** ppUserVirtualBase);

extern DfMmappableStatus
dfVxwMunmapRegion(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj,
  void* pUserVirtualBase);

extern DfMmappableStatus
dfVxwMunmapAllRegions(
	DfDeviceObject* pDevObj,
	DfClientObject* pClObj);

extern void
dfVxwIsrStub(
  DfInterruptObject* pInterruptObject);

extern void
dfVxwCheckDataTypes(
  void);

extern STATUS
dfVxwLegacyEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware);

extern void
dfVxwLegacyExit(
  void);

#endif
